import json
from json import JSONEncoder


class PomeEncoder(JSONEncoder):
    """The json encoder used by pome.
    Keys are sorted and indent is applied in order to have human readable and diffable documents.
    """

    def __init__(self, **kwargs):
        super(PomeEncoder, self).__init__(**kwargs, indent=2)

    def default(self, o):
        return o.__dict__


class PomeEncodable(object):
    @classmethod
    def from_json_dict(cls, json_dict):
        to_return = cls(**json_dict)
        to_return._post_load_json()
        return to_return

    @classmethod
    def from_json_file(cls, filename, take_default_if_no_file_found=False):
        try:
            with open(filename, "r") as f:
                return cls.from_json_dict(json.loads(f.read()))
        except FileNotFoundError:
            if not take_default_if_no_file_found:
                return None
            return cls()

    @classmethod
    def from_disk(cls, take_default_if_no_file_found=False):
        return cls.from_json_file(cls.default_filename, take_default_if_no_file_found)

    def _post_load_json(self):
        """Function to execute once the object has been de-serialised from json."""
        pass

    def to_json(self):
        return PomeEncoder().encode(self)

    def __str__(self):
        return self.to_json()



from pome.models.encoder import PomeEncodable


class Settings(PomeEncodable):

    default_filename = "pome_settings.json"

    def __init__(self, git_communicate_with_remote: bool = True):
        self.git_communicate_with_remote: bool = git_communicate_with_remote



from pome import app


@app.route("/test/add-random-tx", methods=["GET"])
@app.route("/test/add-random-tx/<int:nb_tx>", methods=["GET"])
def add_random_simple_transactions(nb_tx=1):
    import random
    import time
    from datetime import datetime, timedelta

    import requests
    from lorem_text import lorem

    def str_time_prop(start, end, time_format, prop):
        """Get a time at a proportion of a range of two formatted times.

        start and end should be strings specifying times formatted in the
        given format (strftime-style), giving an interval [start, end].
        prop specifies how a proportion of the interval to be taken after
        start.  The returned time will be in the specified format.
        """

        stime = time.mktime(time.strptime(start, time_format))
        etime = time.mktime(time.strptime(end, time_format))

        ptime = stime + prop * (etime - stime)

        return time.strftime(time_format, time.localtime(ptime))

    def random_date(start, end, prop):
        return str_time_prop(start, end, "%Y-%m-%d", prop)

    for _ in range(nb_tx):
        amount = random.randint(0, 100000)
        payload = {}
        if random.randint(0, 1) == 1:
            tx_line = {
                "account_dr": "2000",
                "account_cr": "6003",
                "raw_amount_in_main_currency": str(amount / 100),
            }
        else:
            tx_line = {
                "account_dr": "6003",
                "account_cr": "0400",
                "raw_amount_in_main_currency": str(amount / 100),
            }
        payload["lines"] = [tx_line]
        yesterday = datetime.now() - timedelta(1)
        tx_date = random_date(
            "2021-01-01", datetime.strftime(yesterday, "%Y-%m-%d"), random.random()
        )
        payload["date"] = tx_date
        record_date = random_date(
            tx_date, datetime.strftime(yesterday, "%Y-%m-%d"), random.random()
        )
        payload["date_recorded"] = record_date + "T01:45:36.123Z"

        narrative = ""
        if random.randint(0, 1) == 1:
            narrative = lorem.words(13)
        payload["narrative"] = narrative

        r = requests.post("http://localhost:5000/transactions/record", json=payload)

    return "ok"



import os
import pathlib
from glob import glob


def get_longest_matching_prefix(word, prefix_list):
    """Find the longest prefix of word that is in prefix_list. None otherwise."""
    to_return = None
    for i in range(len(word)):
        prefix = word[: i + 1]
        if prefix in prefix_list:
            to_return = prefix
    return to_return


def get_recursive_json_hash():
    """Gives compound hash of all json stats in repository."""
    the_hash = None
    for dirpath, dirnames, filenames in os.walk("."):
        if ".git" in dirnames:
            dirnames.remove(".git")
        for filetype in ["*.json", "*.csv"]:
            for filename in glob(os.path.join(dirpath, filetype)):
                if os.path.exists(filename):
                    if the_hash is None:
                        the_hash = hash(pathlib.Path(filename).stat())
                    else:
                        the_hash = hash((the_hash, hash(pathlib.Path(filename).stat())))

    return the_hash



from pome.models.encoder import PomeEncodable


class Address(PomeEncodable):
    """Representation of a postal address."""

    def __init__(
        self,
        line_1: str = "",
        line_2: str = "",
        line_3: str = "",
        city: str = "",
        postal_code: str = "",
        country_code: str = "",
    ):
        self.line_1: str = line_1
        self.line_2: str = line_2
        self.line_3: str = line_3
        self.city: str = city
        self.postal_code: str = postal_code
        self.country_code: str = country_code



import os
from typing import Dict, Set, Union

from flask import Flask

from pome._version import __version__
from pome.models import AccountsChart, Company, Settings

app = Flask(__name__)
app.secret_key = b'\x9dq\x0bbE\xbaA{\xb4V`\xcaq\xb7\xcf"\x8b\xb8q\xe0\x13b\xb0\xb6'
app.config["TEMPLATES_AUTO_RELOAD"] = True  # not working

app.jinja_env.globals["POME_VERSION"] = __version__

from git import Git, GitCommandError, InvalidGitRepositoryError, Repo

app.jinja_env.globals["GIT_OK"] = True
app.jinja_env.globals["GIT_PULL_ERROR"] = ""
git: Union[None, Git] = None
try:
    repo = Repo(".")
    git = repo.git
except InvalidGitRepositoryError:
    app.jinja_env.globals["GIT_OK"] = False
    git = None
app.jinja_env.globals["CWD"] = os.getcwd()


class GlobalState(object):
    def __init__(self):
        self.settings: Union[Settings, None] = None
        self.company: Union[Company, None] = None
        self.accounts_chart: Union[AccountsChart, None] = None
        self.recorded_transactions: Union[Dict[str, "Transaction"], None] = None

    def sync_settings_from_disk(self):
        """Settings need to be load before everything else in order to interact with git."""
        self.settings = Settings.from_disk(True)

    def sync_from_disk(self):
        # Avoiding circular imports
        from pome.models.transaction import Transaction

        self.settings = Settings.from_disk(True)

        self.company = Company.from_disk()
        app.jinja_env.globals["company"] = self.company

        self.accounts_chart = AccountsChart.from_disk()
        app.jinja_env.globals["accounts_chart"] = self.accounts_chart

        self.recorded_transactions = Transaction.fetch_all_recorded_transactions()


g = GlobalState()
g.sync_settings_from_disk()


def global_pull():
    try:
        print("Git pull")
        return git.pull()
    except GitCommandError as e:
        app.jinja_env.globals["GIT_PULL_ERROR"] = e.stderr


if git is not None and g.settings.git_communicate_with_remote:
    global_pull()

print("Sync state from disk")
g.sync_from_disk()

from pome.currency import (CURRENCY_SYMBOL, DECIMAL_PRECISION_FOR_CURRENCY,
                           EXAMPLE_MONEY_INPUT)

app.jinja_env.globals["CURRENCY_SYMBOL"] = CURRENCY_SYMBOL
app.jinja_env.globals["EXAMPLE_MONEY_INPUT"] = EXAMPLE_MONEY_INPUT
app.jinja_env.globals["DECIMAL_PRECISION_FOR_CURRENCY"] = DECIMAL_PRECISION_FOR_CURRENCY

import pome.routes
import pome.test_routes



from typing import Dict, List, Tuple, Union

from money.currency import Currency
from money.money import Money

from pome.misc import get_longest_matching_prefix
from pome.models.encoder import PomeEncodable


class BankDetails(PomeEncodable):
    def __init__(
        self,
        bank: str = "",
        IBAN: str = "",
        BIC: str = "",
        account_number: str = "",
        sort_code: str = "",
    ):
        self.bank: str = bank
        self.IBAN: str = IBAN
        self.BIC: str = BIC
        self.account_number: str = account_number
        self.sort_code: str = sort_code


class BankAccountDetails(PomeEncodable):
    def __init__(self, code: str = "", bank_details: BankDetails = BankDetails()):
        self.code: str = code
        self.bank_details: BankDetails = bank_details

    def _post_load_json(self):
        self.bank_details = BankDetails.from_json_dict(self.bank_details)


class Account(PomeEncodable):

    ACCOUNT_TYPES = [
        "INCOME",
        "EXPENSE",
        "ASSET",
        "LIABILITY",
        "EQUITY",
        "ASSET_OR_LIABILITY",
    ]

    def __init__(
        self,
        code: str = "",
        name: str = "",
        type: str = "",
        bank_account_details: Union[BankAccountDetails, None] = None,
    ):
        self.code: str = code
        self.name: str = name

        # Valid types are specified in Account.ACCOUNT_TYPES
        self.type: str = type

        self.bank_account_details = bank_account_details

    def pretty_name(self) -> str:
        return self.code + " - " + self.name

    def is_used(self) -> bool:
        return len(self.transactions_lines()) > 0

    def transactions_lines(self, filter=""):
        """Set filter to 'DR' or 'CR' to get only those type of transactions."""
        from pome import g
        from pome.models.transaction import Transaction, TransactionLine

        to_return: List[Tuple[Transaction, TransactionLine]] = []
        for tx_id in g.recorded_transactions:
            tx: Transaction = g.recorded_transactions[tx_id]
            for line in tx.lines:
                if line.account_dr_code == self.code:
                    if filter == "":
                        to_return.append((tx, line))
                    elif filter == "DR":
                        to_return.append((tx, line))
                if line.account_cr_code == self.code:
                    if filter == "":
                        to_return.append((tx, line))
                    elif filter == "CR":
                        to_return.append((tx, line))

        return sorted(to_return, key=lambda x: x[0].id)

    def balance(
        self, formatted=False, algebrised=False
    ) -> Union[Money, str, Tuple[Money, str]]:
        from pome import g
        from pome.models.transaction import Transaction

        sum_dr = Money("0", Currency(g.company.accounts_currency_code))
        sum_cr = Money("0", Currency(g.company.accounts_currency_code))
        for tx_id in g.recorded_transactions:
            tx: Transaction = g.recorded_transactions[tx_id]
            for line in tx.lines:
                if line.account_dr_code == self.code:
                    sum_dr += line.amount.amount()
                if line.account_cr_code == self.code:
                    sum_cr += line.amount.amount()

        balance = None
        if algebrised:
            if self.type in ["INCOME", "LIABILITY", "EQUITY"]:
                balance = sum_cr - sum_dr
            else:
                balance = sum_dr - sum_cr
        else:
            if sum_cr >= sum_dr:
                winning_side = "CR"
                balance = sum_cr - sum_dr
            else:
                winning_side = "DR"
                balance = sum_dr - sum_cr

        if not formatted:
            if algebrised:
                return balance
            else:
                return balance, winning_side

        if algebrised:
            return balance.format(g.company.locale)
        else:
            return balance.format(g.company.locale) + " " + winning_side

    def _post_load_json(self):
        if self.bank_account_details is not None:
            self.bank_account_details = BankAccountDetails.from_json_dict(
                self.bank_account_details
            )


class AccountsChartSection(PomeEncodable):
    def __init__(self, prefix: str = "", name: str = ""):
        self.prefix: str = prefix
        self.name: str = name


class AccountsChart(PomeEncodable):

    default_filename = "accounts_chart.json"

    def __init__(
        self,
        sections: List[AccountsChartSection] = [],
        accounts_csv_file: Union[None, str] = None,
        accounts: List[Account] = [],
        bank_accounts_details: List[BankAccountDetails] = [],
    ):
        self.sections: List[AccountsChartSection] = sections
        self.accounts_csv_file: Union[None, str] = accounts_csv_file
        self.accounts: List[Account] = accounts
        self.bank_accounts_details: List[BankAccountDetails] = bank_accounts_details

        self.account_codes = None
        pass

    def are_all_accounts_used(self) -> bool:
        for acc in self.accounts:
            if not acc.is_used():
                return False
        return True

    def at_least_one_account_used_in_section(self, section_prefix: str) -> bool:
        if section_prefix not in self.section_account_code_map:
            return False
        for acc in self.section_account_code_map[section_prefix]:
            if self.account_codes[acc].is_used() == True:
                return True
        return False

    def is_valid_account_code(self, code: str):
        if self.account_codes is None:
            raise ValueError("Account codes map is not set in account chart object.")
        return code in self.account_codes

    def _load_accounts_from_csv_file(self, csv_file: str):
        try:
            with open(csv_file, "r") as f:
                csv_content = f.read()

            self.accounts = []

            for csv_line in csv_content.split("\n"):
                if csv_line.strip() == "":
                    continue
                csv_entries = csv_line.split(";")
                if len(csv_entries) != 3:
                    # TODO: make error appear on frontend?
                    print(f"Account csv entry invalid: {csv_line}. Ignored.")
                    continue
                code, name, type_ = list(map(str.strip, csv_entries))
                self.accounts.append(Account(code, name, type_))

        except FileNotFoundError:
            self.accounts_csv_file_error = True
            print(f"Accounts file not found! `{csv_file}`")

    def _make_acounts_code_map(self):
        self.account_codes: Dict[str, Account] = {}
        for acc in self.accounts:
            if acc.code not in self.account_codes:
                self.account_codes[acc.code] = acc
            else:
                # TODO: make error appear on frontend
                print(
                    f"Warning. Account code {acc.code} is not unique: it is at least shared by:\n {self.account_codes[acc.code]} and {acc}. Pome will only keep {self.account_codes[acc.code]}."
                )

    def _check_accounts_type(self):
        for code in self.account_codes:
            acc = self.account_codes[code]
            if acc.type not in Account.ACCOUNT_TYPES:
                # TODO: make error appear on frontend
                print(
                    f"Warning. Account type {acc.type} for account {acc} is not valid. Valid types are: {Account.ACCOUNT_TYPES}"
                )

    def prefix_level(self, prefix: str) -> int:
        """Returns the level of an account prefix in the account chart tree."""
        level = 0
        for i in range(len(prefix)):
            prefix_of = prefix[: i + 1]
            if prefix_of in self.section_prefixes_map:
                level += 1
        return level - 1

    def _make_section_account_code_map(self):
        self.section_prefixes_map = {}
        for section in self.sections:
            if not section.prefix in self.section_prefixes_map:
                self.section_prefixes_map[section.prefix] = section
            else:
                print(
                    f"Warning. Two sections shares the same prefix {section.prefix}, only the first one in order will be considered."
                )

        self.section_account_code_map = {}
        for code in self.account_codes:
            acc = self.account_codes[code]
            longest_matching_prefix = get_longest_matching_prefix(
                code, self.section_prefixes_map.keys()
            )
            if longest_matching_prefix is None:
                print(
                    f"Warning. Account {acc} belongs to no section, it wont be printed."
                )
                continue
            if not longest_matching_prefix in self.section_account_code_map:
                self.section_account_code_map[longest_matching_prefix] = []
            self.section_account_code_map[longest_matching_prefix].append(acc.code)

        for section_prefix in self.section_account_code_map:
            self.section_account_code_map[section_prefix].sort()

    def _post_load_json(self):
        self.sections = list(map(AccountsChartSection.from_json_dict, self.sections))
        self.accounts = list(map(Account.from_json_dict, self.accounts))
        self.bank_accounts_details = list(
            map(BankAccountDetails.from_json_dict, self.bank_accounts_details)
        )

        if self.accounts_csv_file is not None:
            self._load_accounts_from_csv_file(self.accounts_csv_file)

        self._make_acounts_code_map()
        self._check_accounts_type()

        self._make_section_account_code_map()

        return



import os
import pathlib
import re
import shutil
import sys

import pkg_resources
import setuptools

if sys.version_info < (3, 9):
    sys.exit("Sorry, Python < 3.9 is not supported")


VERSIONFILE = "pome/_version.py"

with open(VERSIONFILE, "rt") as f:
    verstrline = f.read()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        verstr = mo.group(1)
    else:
        raise RuntimeError(f"Unable to find version string in {VERSIONFILE}.")

with pathlib.Path("requirements.txt").open() as requirements_txt:
    install_requires = [
        str(requirement)
        for requirement in pkg_resources.parse_requirements(requirements_txt)
    ]


if not os.path.exists("scripts"):
    os.makedirs("scripts")
shutil.copyfile("run_pome.py", "scripts/pome")

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="pome",
    version=verstr,
    scripts=["scripts/pome"],
    author="Tristan Stérin",
    author_email="tristan@prgm.dev",
    description="Decentralised, git-based, accounting software.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pome-gr/pome",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
